@@ -25,7 +25,7 @@ from py4web import (
     request,
     response,
 )
-from py4web.core import ErrorStorage, Fixture, Reloader, Session, dumps
+from py4web.core import Fixture, Reloader, Session, dumps, error_logger, safely
 from py4web.utils.factories import ActionFactory
 
 from .diff2kryten import diff2kryten
@@ -36,8 +36,7 @@ FOLDER = os.environ["PY4WEB_APPS_FOLDER"]
 APP_FOLDER = os.path.dirname(__file__)
 T_FOLDER = os.path.join(APP_FOLDER, "translations")
 T = Translator(T_FOLDER)
-error_storage = ErrorStorage()
-db = error_storage.db
+
 session = Session()
 
 
@@ -109,7 +108,7 @@ if MODE in ("demo", "readonly", "full"):
     @action.uses("index.html", session, T)
     def index():
         return dict(
-            languages=dumps(getattr(T.local, 'language', {})),
+            languages=dumps(getattr(T.local, "language", {})),
             mode=MODE,
             user_id=(session.get("user") or {}).get("id"),
         )
@@ -140,7 +139,7 @@ if MODE in ("demo", "readonly", "full"):
     @action("dbadmin")
     @action.uses(Logged(session), "dbadmin.html")
     def dbadmin():
-        return dict(languages=dumps(getattr(T.local, 'language', {})))
+        return dict(languages=dumps(getattr(T.local, "language", {})))
 
     @action("info")
     @session_secured
@@ -290,19 +289,27 @@ if MODE in ("demo", "readonly", "full"):
     @session_secured
     def tickets():
         """Returns most recent tickets grouped by path+error"""
-        tickets = error_storage.get()
-        return {"payload": tickets}
+        tickets = safely(error_logger.database_logger.get) if MODE != "DEMO" else None
+        return {"payload": tickets or []}
 
     @action("clear")
     @session_secured
     def clear_tickets():
-        error_storage.clear()
+        if MODE != "demo":
+            safely(error_logger.databasea_logger.clear)
 
     @action("ticket/<ticket_uuid>")
     @action.uses("ticket.html")
     @session_secured
     def error_ticket(ticket_uuid):
-        return dict(ticket=ErrorStorage().get(ticket_uuid=ticket_uuid))
+        if MODE != "demo":
+            return dict(
+                ticket=safely(
+                    lambda: error_logger.database_logger.get(ticket_uuid=ticket_uuid)
+                )
+            )
+        else:
+            return dict(ticket=None)
 
     @action("rest/<path:path>", method=["GET", "POST", "PUT", "DELETE"])
     @session_secured
@@ -429,7 +436,7 @@ if MODE == "full":
         form = request.json
         # Directory for zipped assets
         assets_dir = os.path.join(os.path.dirname(py4web.__file__), "assets")
-        app_name = form['name']
+        app_name = form["name"]
         target_dir = safe_join(FOLDER, app_name)
         if form["type"] == "minimal":
             source = os.path.join(assets_dir, "py4web.app._minimal.zip")
