@@ -3,29 +3,27 @@
 """PY4WEB - a web framework for rapid development of efficient database driven web applications"""
 
 # Standard modules
-import argparse
+import asyncio
 import cgitb
 import code
 import copy
 import datetime
+import enum
 import functools
-import importlib.util
-import importlib.machinery
-import inspect
-import json
 import http.client
 import http.cookies
+import importlib.machinery
+import importlib.util
+import inspect
+import json
 import linecache
 import logging
 import numbers
 import os
-import getpass
 import pathlib
 import platform
 import re
 import signal
-import site
-import shutil
 import sys
 import threading
 import time
@@ -34,8 +32,9 @@ import types
 import urllib.parse
 import uuid
 import zipfile
-import asyncio
+
 from watchgod import awatch
+
 from . import server_adapters
 
 # Optional web servers for speed
@@ -44,23 +43,15 @@ try:
 except ImportError:
     gunicorn = None
 
-try:
-    from enum import Enum
-except:
-    # for python < 3.4
-    class Enum:
-        pass
-
+import bottle
 
 # Third party modules
 import click
-import bottle
 import jwt  # this is PyJWT
-import yatl
-import pydal
 import pluralize
-from pydal._compat import to_native, to_bytes
+import pydal
 import threadsafevariable
+import yatl
 
 bottle.BaseRequest.MEMFILE_MAX = 16 * 1024 * 1024
 
@@ -138,6 +129,19 @@ def required_folder(*parts):
     return path
 
 
+def safely(func, exceptions=(Exception,), log=False):
+    """
+    runs the funnction and returns True on success,
+    False if one of the exceptions is raised
+    """
+    try:
+        return func()
+    except exceptions as err:
+        if log:
+            logging.warn(str(err))
+        return None
+
+
 ########################################################################################
 # fix request.fullpath for the case of domain mapping to app
 # (request.url will be autofixed, since it is based on request.fullpath)
@@ -266,7 +270,7 @@ def objectify(obj):
     elif hasattr(obj, "xml"):
         return obj.xml()
     elif isinstance(
-        obj, Enum
+        obj, enum.Enum
     ):  # Enum class handled specially to address self reference in __dict__
         return dict(name=obj.name, value=obj.value, __class__=obj.__class__.__name__)
     elif hasattr(obj, "__dict__") and hasattr(obj, "__class__"):
@@ -295,6 +299,9 @@ class Fixture:
     def on_success(self, status):
         pass  # called when a request is successful
 
+    def finalize(self):
+        pass  # called in any case at the end of a request
+
     def transform(
         self, output, shared_data=None
     ):  # transforms the output, for example to apply template
@@ -328,7 +335,7 @@ class DAL(pydal.DAL, Fixture):
 # make sure some variables in pydal are thread safe
 def thread_safe_pydal_patch():
     Field = pydal.DAL.Field
-    tsafe_attrs = ["readable", "writable", "default", "update", "requires"]
+    tsafe_attrs = ["readable", "writable", "default", "update", "requires", "widget"]
     for a in tsafe_attrs:
         setattr(Field, a, threadsafevariable.ThreadSafeVariable())
 
@@ -357,6 +364,58 @@ thread_safe_pydal_patch()
 # this global object will be used to store their state to restore it for every http request
 ICECUBE = {}
 
+
+#########################################################################################
+# Current Fixture
+#########################################################################################
+
+
+class NotInCurrent(Exception):
+    """This exception is raised when one tries to access a request-local
+    object but one is not in a request-local context."""
+
+    pass
+
+
+class Current(Fixture):
+    """
+    This fixture gives access to a request-local object, that is cleaned
+    after each request.  Note that the object is thread-local; if the
+    request processing uses multiple threads, this will not be accessible.
+    """
+
+    def __init__(self):
+        self._local = threading.local()
+        self.local = None
+
+    def on_request(self):
+        self._local = self._local
+        self.local.data = {}
+
+    def finalize(self):
+        self.local = None
+
+    def __setitem__(self, key, value):
+        if self.local is None:
+            raise NotInCurrent()
+        self.local.data[key] = value
+
+    def __getitem__(self, key):
+        if self.local is None:
+            raise NotInCurrent()
+        return self.local.data[key]
+
+    def __delitem__(self, key):
+        if self.local is None:
+            raise NotInCurrent()
+        del self.local.data[key]
+
+    def get(self, key, default=None):
+        if self.local is None:
+            raise NotInCurrent()
+        return self.local.data.get(key, default)
+
+
 #########################################################################################
 # Flash Fixture
 #########################################################################################
@@ -376,39 +435,45 @@ class Flash(Fixture):
     Also notice all Flash objects share the same threading local so act as singletons
     """
 
+    # this essential makes flash a singleton
+    # necessary because auth defines its own flash
+    # possible because flash does not depend on the app
     local = threading.local()
 
     def on_request(self):
         # when a new request arrives we look for a flash message in the cookie
         flash = request.get_cookie("py4web-flash")
         if flash:
-            Flash.local.flash = json.loads(flash)
+            self.local.flash = json.loads(flash)
         else:
-            Flash.local.flash = None
+            self.local.flash = None
 
     def on_success(self, status):
         # if we redirect and have a flash message we move it to the session
-        if status == 303 and Flash.local.flash:
-            response.set_cookie("py4web-flash", json.dumps(Flash.local.flash), path="/")
-            Flash.local.flash = None
+        if status == 303 and self.local.flash:
+            response.set_cookie("py4web-flash", json.dumps(self.local.flash), path="/")
         else:
             response.delete_cookie("py4web-flash", path="/")
 
+    def finalize(self):
+        """Clears the local to prevent leakage."""
+        self.local.__dict__.clear()
+
     def set(self, message, _class="", sanitize=True):
         # we set a flash message
         if sanitize:
             message = yatl.sanitizer.xmlescape(message)
-        Flash.local.flash = {"message": message, "class": _class}
+        self.local.flash = {"message": message, "class": _class}
 
     def transform(self, data, shared_data=None):
         # if we have a valid flash message, we inject it in the response dict
         if isinstance(data, dict):
             if not "flash" in data:
-                data["flash"] = Flash.local.flash or ""
+                data["flash"] = self.local.flash or ""
         else:
-            if Flash.local.flash is not None:
-                response.headers["component-flash"] = json.dumps(Flash.local.flash)
-        Flash.local.flash = None
+            if self.local.flash is not None:
+                response.headers["component-flash"] = json.dumps(self.local.flash)
+        self.local.flash = None
         return data
 
 
@@ -495,22 +560,28 @@ class Session(Fixture):
         self.secret = secret or Session.SECRET
         self.expiration = expiration
         self.algorithm = algorithm
-        self.local = threading.local()
-        self.local.changed = False
-        self.local.secure = None
-        self.local.data = {}
         self.storage = storage
         self.same_site = same_site
         if isinstance(storage, Session):
             self.__prerequisites__ = [storage]
         if hasattr(storage, "__prerequisites__"):
             self.__prerequisites__ = storage.__prerequisites__
+        self._local = threading.local()
+        self.local = None  # We initialize this per-request.
+
+    def initialize(self, app_name="unknown", data=None, changed=False, secure=False):
+        self.local = self._local
+        self.local.changed = changed
+        self.local.data = data or {}
+        self.local.session_cookie_name = "%s_session" % app_name
+        self.local.secure = secure
 
     def load(self):
-        self.local.session_cookie_name = "%s_session" % request.app_name
-        self.local.changed = False
-        self.local.secure = request.url.startswith("https")
-        self.local.data = {}
+        self.initialize(
+            app_name=request.app_name,
+            changed=False,
+            secure=request.url.startswith("https"),
+        )
         raw_token = request.get_cookie(
             self.local.session_cookie_name
         ) or request.query.get("_session_token")
@@ -519,7 +590,7 @@ class Session(Fixture):
                 request.json and request.json.get("_session_token")
             )
         if raw_token:
-            token_data = to_bytes(raw_token)
+            token_data = raw_token.encode()
             try:
                 if self.storage:
                     json_data = self.storage.get(token_data)
@@ -551,10 +622,12 @@ class Session(Fixture):
             cookie_data = jwt.encode(
                 self.local.data, self.secret, algorithm=self.algorithm
             )
+            if isinstance(cookie_data, bytes):
+                cookie_data = cookie_data.decode()
 
         response.set_cookie(
             self.local.session_cookie_name,
-            to_native(cookie_data),
+            cookie_data,
             path="/",
             secure=self.local.secure,
             same_site=self.same_site,
@@ -583,6 +656,7 @@ class Session(Fixture):
             yield item
 
     def clear(self):
+        """Produces a brand-new session."""
         self.local.changed = True
         self.local.data.clear()
         self.local.data["uuid"] = str(uuid.uuid1())
@@ -599,6 +673,9 @@ class Session(Fixture):
         if self.local.changed:
             self.save()
 
+    def finalize(self):
+        self.local = None  # To prevent leakage
+
 
 #########################################################################################
 # The URL Helper
@@ -709,7 +786,6 @@ def redirect(location):
 class action:
     """@action(...) is a decorator for functions to be exposed as actions"""
 
-    current = threading.local()
     registered = set()
     app_name = "_default"
 
@@ -757,6 +833,9 @@ class action:
                 except Exception:
                     [obj.on_error() for obj in fixtures]
                     raise
+                finally:
+                    [obj.finalize() for obj in fixtures]
+                    # Clears the current object to prevent leakage.
 
             return wrapper
 
@@ -797,15 +876,14 @@ class action:
             except bottle.HTTPResponse:
                 raise
             except Exception:
-                logging.error(traceback.format_exc())
-                try:
-                    ticket = ErrorStorage().log(request.app_name, get_error_snapshot())
-                except Exception:
-                    logging.error(traceback.format_exc())
-                    ticket = "unknown"
+                snapshot = get_error_snapshot()
+                logging.error(snapshot["traceback"])
+                ticket_uuid = error_logger.log(request.app_name, snapshot) or "unknown"
                 raise bottle.HTTPResponse(
                     body=error_page(
-                        500, button_text=ticket, href="/_dashboard/ticket/" + ticket
+                        500,
+                        button_text=ticket_uuid,
+                        href="/_dashboard/ticket/" + ticket_uuid,
                     ),
                     status=500,
                 )
@@ -937,8 +1015,16 @@ def get_error_snapshot(depth=5):
     return data
 
 
-class ErrorStorage:
+class SimpleErrorLogger:
+    def log(self, app_name, snapshot):
+        """logs the error"""
+        logging.error("%s error:\n%s" % (app_name, snapshot["traceback"]))
+        return None
+
+
+class DatabaseErrorLogger:
     def __init__(self):
+        """creates the py4web_error table in the service database"""
         uri = os.environ["PY4WEB_SERVICE_DB_URI"]
         folder = os.environ["PY4WEB_SERVICE_FOLDER"]
         self.db = DAL(uri, folder=folder)
@@ -956,6 +1042,7 @@ class ErrorStorage:
         self.db.commit()
 
     def log(self, app_name, error_snapshot):
+        """store error snapshot (ticket) in the database"""
         ticket_uuid = str(uuid.uuid4())
         try:
             id = self.db.py4web_error.insert(
@@ -970,11 +1057,13 @@ class ErrorStorage:
             )
             self.db.commit()
             return ticket_uuid
-        except Exception:
+        except Exception as err:
+            logging.error(str(err))
             self.db.rollback()
-            return "internal-error"
+            return None
 
     def get(self, ticket_uuid=None):
+        """retrieve a ticket from error database"""
         db = self.db
         if ticket_uuid:
             query, orderby = db.py4web_error.uuid == ticket_uuid, None
@@ -999,11 +1088,51 @@ class ErrorStorage:
         return rows if not ticket_uuid else rows[0] if rows else None
 
     def clear(self):
+        """erase all tickets from database"""
         db = self.db
         db(db.py4web_error).delete()
         self.db.commit()
 
 
+class ErrorLogger:
+
+    """
+    To create your own custom logger for an app:
+
+    class MyLogger:
+        def log(app_name, error_snap_shop):
+            ...
+            return ticket_uuid
+    
+    error_logger.plugins['app_name'] = MyLogger()
+    """
+
+    def __init__(self):
+        self.fallback_logger = SimpleErrorLogger()
+        self.database_logger = None
+        self.plugins = {}
+
+    def initialize(self):
+        """try inizalize database if we have service folder"""
+        self.database_logger = safely(DatabaseErrorLogger, log=True)
+
+    def _get_logger(self, app_name):
+        """get the appropriate logger for the app"""
+        return (
+            self.plugins.get(app_name) or self.database_logger or self.fallback_logger
+        )
+
+    def log(self, app_name, error_snapshot):
+        """log the error snapshot"""
+        logger = self._get_logger(app_name)
+        ticket_uuid = safely(lambda: logger.log(app_name, error_snapshot))
+        if not ticket_uuid:
+            self.fallback_logger.log(app_name, error_snapshot)
+        return ticket_uuid
+
+
+error_logger = ErrorLogger()
+
 #########################################################################################
 # Loading &  Reloading Logic
 #########################################################################################
@@ -1091,14 +1220,13 @@ class Reloader:
                 click.secho("\x1b[A[X] loaded %s       " % app_name, fg="green")
                 Reloader.MODULES[app_name] = module
                 Reloader.ERRORS[app_name] = None
-            except:
-                tb = traceback.format_exc()
-                print(tb)
+            except Exception as err:
+                Reloader.ERRORS[app_name] = traceback.format_exc()
+                error_logger.log(app_name, get_error_snapshot())
                 click.secho(
-                    "\x1b[A[FAILED] loading %s       \n%s\n" % (app_name, tb),
+                    "\x1b[A[FAILED] loading %s (%s)" % (app_name, err),
                     fg="red",
                 )
-                Reloader.ERRORS[app_name] = tb
                 # clear all files/submodules if the loading fails
                 clear_modules()
                 return None
@@ -1144,6 +1272,9 @@ class Reloader:
 # Web Server and Reload Logic: Error Handling
 #########################################################################################
 
+ERROR_PAGES = {
+    "*": '<html><head><style>body{color:white;text-align: center;background-color:[[=color]];font-family:serif} h1{font-size:6em;margin:16vh 0 8vh 0} h2{font-size:2em;margin:8vh 0} a{color:white;text-decoration:none;font-weight:bold;padding:10px 10px;border-radius:10px;border:2px solid #fff;transition: all .5s ease} a:hover{background:rgba(0,0,0,0.1);padding:10px 30px}</style></head><body><h1>[[=code]]</h1><h2>[[=message]]</h2>[[if button_text:]]<a href="[[=href]]">[[=button_text]]</a>[[pass]]</body></html>',
+}
 
 def error_page(code, button_text=None, href="#", color=None, message=None):
     message = http.client.responses[code].upper() if message is None else message
@@ -1160,11 +1291,8 @@ def error_page(code, button_text=None, href="#", color=None, message=None):
         response.status = code
         return json.dumps(context)
     # else - return html error-page
-    return yatl.render(
-        '<html><head><style>body{color:white;text-align: center;background-color:[[=color]];font-family:serif} h1{font-size:6em;margin:16vh 0 8vh 0} h2{font-size:2em;margin:8vh 0} a{color:white;text-decoration:none;font-weight:bold;padding:10px 10px;border-radius:10px;border:2px solid #fff;transition: all .5s ease} a:hover{background:rgba(0,0,0,0.1);padding:10px 30px}</style></head><body><h1>[[=code]]</h1><h2>[[=message]]</h2>[[if button_text:]]<a href="[[=href]]">[[=button_text]]</a>[[pass]]</body></html>',
-        context=context,
-        delimiters="[[ ]]",
-    )
+    content = ERROR_PAGES.get(code) or ERROR_PAGES["*"]
+    return yatl.render(content, context=context, delimiters="[[ ]]")
 
 
 @bottle.error(404)
@@ -1193,8 +1321,8 @@ def error404(error):
 
 DIRTY_APPS = dict()  #  apps that need to be reloaded (lazy watching)
 
-from inspect import stack
 from collections import OrderedDict
+from inspect import stack
 
 APP_WATCH = {"files": dict(), "handlers": OrderedDict(), "tasks": dict()}
 
@@ -1431,6 +1559,19 @@ def install_args(kwargs, reinstall_apps=False):
     if apps_folder_name != "apps":
         MetaPathRouter(apps_folder_name)
 
+    if not os.path.exists(kwargs["service_folder"]):
+        os.mkdir(kwargs["service_folder"])
+    session_secret_filename = os.path.join(kwargs["service_folder"], "session.secret")
+    if not os.path.exists(session_secret_filename):
+        with open(session_secret_filename, "w") as fp:
+            fp.write(str(uuid.uuid4()))
+
+    with open(session_secret_filename) as fp:
+        Session.SECRET = fp.read()
+
+    # after everything is etup but before installing apps, init
+    error_logger.initialize()
+
     # Reinstall apps from zipped ones in assets
     if reinstall_apps:
         assets_dir = os.path.join(os.path.dirname(__file__), "assets")
@@ -1450,16 +1591,6 @@ def install_args(kwargs, reinstall_apps=False):
                             zip_file.extractall(target_dir)
                             click.echo("\x1b[A[X]")
 
-    if not os.path.exists(kwargs["service_folder"]):
-        os.mkdir(kwargs["service_folder"])
-    session_secret_filename = os.path.join(kwargs["service_folder"], "session.secret")
-    if not os.path.exists(session_secret_filename):
-        with open(session_secret_filename, "w") as fp:
-            fp.write(str(uuid.uuid4()))
-
-    with open(session_secret_filename) as fp:
-        Session.SECRET = fp.read()
-
 
 def wsgi(**kwargs):
     """Initializes everything, loads apps, returns the wsgi app"""
