@@ -24,7 +24,7 @@ from yatl.helpers import (
 )
 from pydal.objects import Field, FieldVirtual
 from py4web import request, URL, response, redirect, HTTP
-from py4web.utils.form import Form, FormStyleDefault
+from py4web.utils.form import Form, FormStyleDefault, join_classes
 from py4web.utils.param import Param
 
 NAV = TAG.nav
@@ -162,6 +162,7 @@ class GridClassStyleBulma(GridClassStyle):
         "grid-tr": "",
         "grid-th": "",
         "grid-td": "",
+        "grid-button": "grid-button button is-small",
         "grid-details-button": "grid-details-button button is-small",
         "grid-edit-button": "grid-edit-button button is-small",
         "grid-delete-button": "grid-delete-button button is-small",
@@ -233,10 +234,24 @@ class GridClassStyleBulma(GridClassStyle):
     }
 
 
+class Column:
+    """class used to represent a column in a grid"""
+
+    def __init__(self, name, represent):
+        self.name = name
+        self.represent = represent
+
+    def render(self, row, index=None):
+        """renders a row al position index (optional)"""
+        return self.represent(row)
+
+
 class Grid:
 
     FORMATTERS_BY_TYPE = {
-        "boolean": lambda value: INPUT(_type="checkbox", _checked=value)
+        "boolean": lambda value: INPUT(
+            _type="checkbox", _checked=value, _disabled="disabled"
+        )
         if value
         else "",
         "datetime": lambda value: XML(
@@ -279,7 +294,7 @@ class Grid:
         query,
         search_form=None,
         search_queries=None,
-        fields=None,
+        columns=None,
         field_id=None,
         show_id=False,
         orderby=None,
@@ -297,6 +312,8 @@ class Grid:
         search_button_text="Filter",
         formstyle=FormStyleDefault,
         grid_class_style=GridClassStyle,
+        # deprecated
+        fields=None,
     ):
         """
         Grid is a searchable/sortable/pageable grid
@@ -305,7 +322,7 @@ class Grid:
         :param query: the query used to filter the data
         :param search_form: py4web FORM to be included as the search form
         :param search_queries: future use - pass a dict of name and a search query
-        :param fields: list of fields to display on the list page, if blank, glean tablename from first query
+        :param columns: list of columns to display on the list page, if blank, glean tablename from first query
         :              and use all fields of that table
         :param field_id: the id field of the primary table for the grid - if there are multiple tables (joined)
                             then the table for this field is used as the table for edit/details/delete
@@ -330,7 +347,7 @@ class Grid:
         self.db = query._db
         self.param = Param(
             query=query,
-            fields=fields,
+            columns=columns or fields,
             field_id=field_id,
             show_id=show_id,
             orderby=orderby,
@@ -373,7 +390,6 @@ class Grid:
         self.page_end = None
         self.page_start = None
         self.query_parms = request.params
-        self.readonly_fields = None
         self.record_id = None
         self.rows = None
         self.tablename = None
@@ -412,16 +428,32 @@ class Grid:
             self.tablename = self.get_tablenames(self.param.query)[0]
         self.record_id = safe_int(parts[1] if len(parts) > 1 else None, default=None)
 
-        if self.param.fields:
-            if not isinstance(self.param.fields, list):
-                self.param.fields = [self.param.fields]
+        table = db[self.tablename]
+        if not self.param.columns:
+            # if no column specified use all fields
+            self.param.columns = [field for field in table if field.readable]
+            self.needed_fields = self.param.columns[:]
+        elif any(isinstance(col, Column) for col in self.param.columns):
+            # if we use columns we have to get all fields and assume a single table
+            self.needed_fields = [field for field in db[self.tablename]]
+        elif any(isinstance(col, FieldVirtual) for col in self.param.columns):
+            # if virtual fields are specified the fields may come from a join
+            needed_fields = set()
+            for col in self.param.columns:
+                if isinstance(col, Field):
+                    needed_fields.add(col)
+                elif isinstance(col, FieldVirtual):
+                    for field in db[col.tablename]:
+                        needed_fields.add(field)
+            self.needed_fields = list(needed_fields)
         else:
-            table = db[self.tablename]
-            self.param.fields = [field for field in table if field.readable]
+            # the columns specify with fields are needed
+            self.needed_fields = self.param.columns[:]
+
+        # except the primary key may be missing and must be fetched even if not displayed
+        if not any(col.name == table._id.name for col in self.needed_fields):
+            self.needed_fields.insert(0, table._id)
 
-        self.readonly_fields = [
-            field for field in self.param.fields if not field.writable
-        ]
         self.referrer = None
 
         if not self.tablename:
@@ -440,8 +472,6 @@ class Grid:
                     redirect(self.endpoint)  ## maybe flash
 
             readonly = self.action == "details"
-            for field in self.readonly_fields:
-                db[field.tablename][field.name].writable = False
 
             if not self.param.show_id:
                 #  if not show id, find the "id" field and set readable/writable to False
@@ -455,6 +485,7 @@ class Grid:
                 db[self.tablename],
                 record=self.record_id,
                 readonly=readonly,
+                deletable=self.param.deletable,
                 formstyle=self.param.formstyle,
                 **attrs,
             )
@@ -504,16 +535,16 @@ class Grid:
             #  find the primary key of the primary table
             pt = db[self.tablename]
             key_is_missing = True
-            for field in self.param.fields:
+            for field in self.param.columns:
                 if (
-                    not isinstance(field, FieldVirtual)
+                    isinstance(field, Field)
                     and field.table._tablename == pt._tablename
                     and field.name == pt._id.name
                 ):
                     key_is_missing = False
             if key_is_missing:
                 #  primary key wasn't included, add it and set show_id to False so it doesn't display
-                self.param.fields.append(pt._id)
+                self.param.columns.append(pt._id)
                 self.param.show_id = False
 
             self.current_page_number = safe_int(request.query.get("page"), default=1)
@@ -560,13 +591,8 @@ class Grid:
                     self.page_start = 1
                 self.page_end = self.total_number_of_rows
 
-            if self.param.fields:
-                param_fields = [
-                    x for x in self.param.fields if not isinstance(x, FieldVirtual)
-                ]  # don't send virtual fields to DAL select - will calculate them later
-                self.rows = db(query).select(*param_fields, **select_params)
-            else:
-                self.rows = db(query).select(**select_params)
+            # get the data
+            self.rows = db(query).select(*self.needed_fields, **select_params)
 
             self.number_of_pages = self.total_number_of_rows // self.param.rows_per_page
             if self.total_number_of_rows % self.param.rows_per_page > 0:
@@ -727,7 +753,7 @@ class Grid:
                 sb.append(field.label)
                 td.append(sb)
             else:
-                td.append(self.param.search_form.custom["widgets"][field.name])
+                td.append(self.param.search_form.custom["wrappers"][field.name])
             if (
                 field.name in self.param.search_form.custom["errors"]
                 and self.param.search_form.custom["errors"][field.name]
@@ -766,50 +792,28 @@ class Grid:
         return div
 
     def render_table_header(self):
-
-        up = I(**self.param.grid_class_style.get("grid-sorter-icon-up"))
-        dw = I(**self.param.grid_class_style.get("grid-sorter-icon-down"))
         columns = []
         sort_order = request.query.get("orderby", "")
 
-        for index, field in enumerate(self.param.fields):
-            if field.readable and (field.type != "id" or self.param.show_id):
-                key = "%s.%s" % (field.tablename, field.name)
-                heading = (
-                    self.param.headings[index]
-                    if index < len(self.param.headings)
-                    else field.label
-                )
-                heading = title(heading)
-                #  add the sort order query parm
-                sort_query_parms = dict(self.query_parms)
-
-                attrs = {}
-                if isinstance(field, FieldVirtual):
-                    col = DIV(heading)
-                elif key == sort_order:
-                    sort_query_parms["orderby"] = "~" + key
-                    url = URL(self.endpoint, vars=sort_query_parms)
-                    attrs = self.attributes_plugin.link(url=url)
-                    col = A(heading, up, **attrs)
-                else:
-                    sort_query_parms["orderby"] = key
-                    url = URL(self.endpoint, vars=sort_query_parms)
-                    attrs = self.attributes_plugin.link(url=url)
-                    col = A(heading, dw if "~" + key == sort_order else "", **attrs)
-                columns.append((key, col))
-
         thead = THEAD(_class=self.param.grid_class_style.classes.get("grid-thead", ""))
-        for key, col in columns:
-            col_class = " grid-col-%s" % key
-            thead.append(
-                TH(
-                    col,
-                    _class=self.param.grid_class_style.classes.get("grid-th", "")
-                    + col_class,
-                    _style=self.param.grid_class_style.styles.get("grid-th"),
+        for index, column in enumerate(self.param.columns):
+            col = None
+            if isinstance(column, (Field, FieldVirtual)):
+                field = column
+                if field.readable and (field.type != "id" or self.param.show_id):
+                    key, col = self.render_field_header(column, index, sort_order)
+            elif isinstance(column, Column):
+                key = column.name.lower().replace(" ", "-")
+                col = column.name
+            else:
+                raise RuntimeError("Invallid Grid Column type")
+            if col:
+                classes = join_classes(
+                    self.param.grid_class_style.classes.get("grid-th"),
+                    "grid-col-%s" % key,
                 )
-            )
+                style = self.param.grid_class_style.styles.get("grid-th")
+                thead.append(TH(col, _class=classes, _style=style))
 
         if (
             self.param.details
@@ -824,6 +828,35 @@ class Grid:
 
         return thead
 
+    def render_field_header(self, field, field_index, sort_order):
+        up = I(**self.param.grid_class_style.get("grid-sorter-icon-up"))
+        dw = I(**self.param.grid_class_style.get("grid-sorter-icon-down"))
+
+        key = "%s.%s" % (field.tablename, field.name)
+        heading = (
+            self.param.headings[field_index]
+            if field_index < len(self.param.headings)
+            else field.label
+        )
+        heading = title(heading)
+        #  add the sort order query parm
+        sort_query_parms = dict(self.query_parms)
+
+        attrs = {}
+        if isinstance(field, FieldVirtual):
+            col = SPAN(heading)
+        elif key == sort_order:
+            sort_query_parms["orderby"] = "~" + key
+            url = URL(self.endpoint, vars=sort_query_parms)
+            attrs = self.attributes_plugin.link(url=url)
+            col = A(heading, up, **attrs)
+        else:
+            sort_query_parms["orderby"] = key
+            url = URL(self.endpoint, vars=sort_query_parms)
+            attrs = self.attributes_plugin.link(url=url)
+            col = A(heading, dw if "~" + key == sort_order else "", **attrs)
+        return key, col
+
     def render_field(self, row, field, field_index):
         """
         Render a field
@@ -892,15 +925,30 @@ class Grid:
                 row_id = row["id"]
                 self.use_tablename = False
 
+            key = "%s.%s" % (self.tablename, "__row")
+            if self.formatters.get(key):
+                extra_class = self.formatters.get(key)(row)["_class"]
+                extra_style = self.formatters.get(key)(row)["_style"]
+            else:
+                extra_class = ""
+                extra_style = ""
             tr = TR(
                 _role="row",
-                _class=self.param.grid_class_style.classes.get("grid-tr"),
-                _style=self.param.grid_class_style.styles.get("grid-tr"),
+                _class=join_classes(
+                    self.param.grid_class_style.classes.get("grid-tr"), extra_class
+                ),
+                _style=join_classes(
+                    self.param.grid_class_style.styles.get("grid-tr"), extra_style
+                ),
             )
             #  add all the fields to the row
-            for index, field in enumerate(self.param.fields):
-                if field.readable and (field.type != "id" or self.param.show_id):
-                    tr.append(self.render_field(row, field, index))
+            for index, column in enumerate(self.param.columns):
+                if isinstance(column, (Field, FieldVirtual)):
+                    field = column
+                    if field.readable and (field.type != "id" or self.param.show_id):
+                        tr.append(self.render_field(row, field, index))
+                elif isinstance(column, Column):
+                    tr.append(TD(column.render(row, index)))
 
             td = None
 
@@ -1078,7 +1126,9 @@ class Grid:
                     override_classes=self.param.grid_class_style.classes.get(
                         "grid-new-button", ""
                     ),
-                    override_styles=self.param.grid_class_style.get("new_button"),
+                    override_styles=self.param.grid_class_style.styles.get(
+                        "grid-new-button"
+                    ),
                 )
             )
 
